#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>

int* create_rand_nums(int size) {
    int* rand_nums = (int*)malloc(sizeof(int) * size);
    for (int i = 0; i < size; i++) {
        // Generating random numbers between 0 and 1 for demonstration purposes
        rand_nums[i] = (int)rand() % 10;
    }
    return rand_nums;
}

int main(int argc, char** argv) {
    MPI_Init(&argc, &argv);

    int world_rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);

    int world_size;
    MPI_Comm_size(MPI_COMM_WORLD, &world_size);

    // Assuming block_size is defined and initialized
    int in_length = 16;
    int block_size = in_length / world_size;

    int* rand_nums = NULL;
    if (world_rank == 0) {
        rand_nums = create_rand_nums(in_length);
    }

    // Create a buffer that will hold a subset of the random numbers
    int* sub_rand_nums = (int*)malloc(sizeof(int) * block_size);

    int* block_prefix = (int*)malloc(sizeof(int) * block_size);


    // Scatter the random numbers to all processes
    MPI_Scatter(rand_nums, block_size, MPI_INT, sub_rand_nums,
                block_size, MPI_INT, 0, MPI_COMM_WORLD);

    // Compute the average of your subset
    block_prefix[0] = sub_rand_nums[0];
    for(int i = 1; i < block_size; i++){
        block_prefix[i] = block_prefix[i-1] + sub_rand_nums[i];
    }
    int sum = block_prefix[block_size - 1];

    // Gather all partial averages down to the root process
    int* block_sum = NULL;
    if (world_rank == 0) {
        block_sum = (int*)malloc(sizeof(int) * world_size);
    }

    MPI_Gather(&sum, 1, MPI_INT, block_sum, 1, MPI_INT, 0, MPI_COMM_WORLD);

    int* prefix_block_sum = NULL;
    if (world_rank == 0) {
        prefix_block_sum = (int*)malloc(sizeof(int) * world_size);
        prefix_block_sum[0] = 0;
        for(int i = 1; i < world_size; i++){
            prefix_block_sum[i] = prefix_block_sum[i-1] + block_sum[i-1];
        }
    }

    int pre;


    MPI_Scatter(prefix_block_sum, 1, MPI_INT, &pre,
                1, MPI_INT, 0, MPI_COMM_WORLD);

    for(int i = 0; i < block_size; i++){
        block_prefix[i] += pre;
    }

    int* out = NULL;
    if (world_rank == 0) {
        out = (int*)malloc(sizeof(int) * in_length);
    }

     MPI_Gather(block_prefix, block_size, MPI_INT, out, block_size, MPI_INT, 0, MPI_COMM_WORLD);



    // Compute the total average of all numbers.
    if (world_rank == 0) {
	printf("Input Array: ");
        for(int i = 0; i < in_length; i++){
            printf("%d ", rand_nums[i]);
        }
	printf("\nPrefix-Sum Array: ");
	for(int i = 0; i < in_length; i++){
		printf("%d ", out[i]);
	
	}
	printf("\n");
        
    }

    free(rand_nums);
    free(sub_rand_nums);
    free(block_sum);

    MPI_Finalize();

    return 0;
}

