#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mpi.h>
#include <time.h>


int tree_reduce(int *send_data, int *recv_data, int count, MPI_Comm comm) {
  int rank_in, np;
  MPI_Comm_rank(MPI_COMM_WORLD, &rank_in);
  MPI_Comm_size(MPI_COMM_WORLD, &np);
  int rank = rank_in;
  int num = np;
  int depth = 1;
  // many sends during a reduce process
  while(num > 1) {
    if(rank < num) {
     
      if(rank % 2 != 0) {
        MPI_Send(send_data, count, MPI_INT, (rank - 1) * depth, 0, comm);
        rank *= num;
        break; // one process can only send once
      } else {
        if(rank != (num - 1)) {
          MPI_Recv(recv_data, count, MPI_INT, (rank + 1) * depth, 0, comm, MPI_STATUS_IGNORE);

          for (int i = 0; i < count; i++) {
            send_data[i] += recv_data[i];
          }
        }
        rank /= 2;
      }
      depth *= 2;
    }
    num = num / 2 + (num%2);
}
 // last reduce
  if(rank_in == 0) {
    int typesize;
    MPI_Type_size(MPI_INT, &typesize);
    memcpy(recv_data, send_data, count * typesize);
  }
  return 0;
}



void reduce_sequential(
    int* send_data,
    int* recv_data,
    int count,
    MPI_Comm communicator)
{
    int my_rank;
    int com_size;
    MPI_Comm_rank(communicator, &my_rank);
    MPI_Comm_size(communicator, &com_size);

    int* gather_buffer = NULL;
    if (my_rank == 0)
    {
        gather_buffer = (int*) calloc(count * com_size, sizeof(int));
    }

    MPI_Gather(send_data, count, MPI_INT, gather_buffer, count, MPI_INT, 0, communicator);

    if (my_rank == 0)
    {
        memset(recv_data, 0, count * sizeof(int));
        for (int p = 0; p < com_size; p++)
            for (int i = 0; i < count; i++)
                recv_data[i] += gather_buffer[count * p + i];
        free(gather_buffer);
    }
}



int main(int argc, char *argv[])
{
int np, rank;
  MPI_Init(&argc, &argv);

  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &np);

  int i = rank + 1, j = rank + 1, sum_par = 0, sum_seq = 0;

  //MPI_Type_size(ctype, &size);
  clock_t par_start_time = clock();
  tree_reduce(&i, &sum_par, 1, MPI_COMM_WORLD);
  clock_t par_end_time = clock();
  double parallel_time = ((double) (par_end_time - par_start_time)) / CLOCKS_PER_SEC;
  clock_t seq_start_time = clock();
  reduce_sequential(&j, &sum_seq, 1, MPI_COMM_WORLD);
  clock_t seq_end_time = clock();
  double sequential_time = ((double) (seq_end_time - seq_start_time)) / CLOCKS_PER_SEC;

  // output result
  if(rank == 0) {
    printf("reduce_tree is %d with time %f, reduce_sequential is %d with time %f\n",
                    sum_par, parallel_time, sum_seq, sequential_time);

  }
  MPI_Finalize();
  return 0;
}


